// Copyright 2023 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build loong64 && linux

package cpu

// This is initialized by archauxv and should not be changed after it is
// initialized.
var HWCap uint

// HWCAP bits. These are exposed by the Linux kernel.
const (
	hwcap_LOONGARCH_CPUCFG = 1 << 0
	hwcap_LOONGARCH_LAM    = 1 << 1
	hwcap_LOONGARCH_UAL    = 1 << 2
	hwcap_LOONGARCH_FPU    = 1 << 3
	hwcap_LOONGARCH_LSX    = 1 << 4
	hwcap_LOONGARCH_LASX   = 1 << 5
	hwcap_LOONGARCH_CRC32  = 1 << 6
)

// CPUCFG registers
const (
	LOONG64_CPUCFG0 = iota
	LOONG64_CPUCFG1
	LOONG64_CPUCFG2
	LOONG64_CPUCFG3
	LOONG64_CPUCFG4
	LOONG64_CPUCFG5
	LOONG64_CPUCFG6
	LOONG64_CPUCFG7
	LOONG64_CPUCFG8
	LOONG64_CPUCFG9
	LOONG64_CPUCFG10
	LOONG64_CPUCFG11
	LOONG64_CPUCFG12
	LOONG64_CPUCFG13
	LOONG64_CPUCFG14
	LOONG64_CPUCFG15
	LOONG64_CPUCFG16
	LOONG64_CPUCFG17
	LOONG64_CPUCFG18
	LOONG64_CPUCFG19
	LOONG64_CPUCFG20
)

// Bit fields for CPUCFG registers, Related reference documents:
// https://loongson.github.io/LoongArch-Documentation/LoongArch-Vol1-EN.html#_cpucfg
const (
	// CPUCFG1 bits
	CPUCFG1_CRC32 = 1 << 25

	// CPUCFG2 bits
	CPUCFG2_LAM_BH = 1 << 27
	CPUCFG2_LAMCAS = 1 << 28
)

func hwcapInit() {
	// It is not taken from CPUCFG data regardless of availability of
	// CPUCFG, because the CPUCFG data only reflects capabilities of the
	// hardware, but not kernel support.
	//
	// These features only require hardware capability support and do not
	// require kernel specific support, so they can be obtained directly
	// through CPUCFG
	Loong64.HasCRC32 = cpucfgIsSetCRC32()
	Loong64.HasLAMCAS = cpucfgIsSetLAMCAS()
	Loong64.HasLAM_BH = cpucfgIsSetLAM_BH()

	// TODO: Features that require kernel support like LSX and LASX can
	// be detected here once needed in std library or by the compiler.
	Loong64.HasLSX = hwcIsSet(HWCap, hwcap_LOONGARCH_LSX)
}

func hwcIsSet(hwc uint, value uint) bool {
	return hwc&value != 0
}

func cpucfgIsSetCRC32() bool {
	return (get_cpucfg(LOONG64_CPUCFG1) & CPUCFG1_CRC32) != 0
}

func cpucfgIsSetLAMCAS() bool {
	return (get_cpucfg(LOONG64_CPUCFG2) & CPUCFG2_LAMCAS) != 0
}

func cpucfgIsSetLAM_BH() bool {
	return (get_cpucfg(LOONG64_CPUCFG2) & CPUCFG2_LAM_BH) != 0
}
